-------------------------------------------------------------------------------------------------
-- FERNANDO MORAES/NEY CALAZANS
--  Criado em 01/NOV/2006
--  Revisado em 14/Jun/2007
--
-- Perifrico para transmisso serial de dados, comandado a partir de uma interface externa
--
-- Este perifrico possui 6 registradores escritos a partir de ordem de um
--		agente externo (tipicamente um processador). Estes registradores
--		definem:
--	reg 0:		O nmero de palavras a transmitir: entre 1 e 4
--	regs 1 a 4:	As palavras de 32 bits a transmitir
--	reg 5:		Apenas seu bit 0 interessa. Quando em 1, ordena o incio da transmisso
--
-- A leitura deste perifrico pela entidade externa permite acesso a 1 nico bit de 
-- informao (ready), informando que o perifrico est disponvel para transmitir
-- dados, seja porque j foi inicializado, seja porque uma transmisso solicitada
-- anteriormente foi concluda.
--
-------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
     -----
     ----- A INTERFACE EXTERNA  SEMELHANTE  INTERFACE EXTERNA DO PROCESSADOR MIPS
     -----

entity transmissor is
    port( clock, reset:  in    std_logic;
          ce, rw:        in    std_logic;
          endereco:      in    std_logic_vector(2 downto 0);   -- 3 bits apenas de endereamento
          data:          inout std_logic_vector(31 downto 0);
          linha:         out   std_logic                       -- sada do perifrico (1 bit)
        );
end transmissor;

architecture transmissor of transmissor is
    type STATES is (REP, INICIO, FIRST, TX, LAST);
    signal EA, PE : STATES;  
    
    -- contadores
    signal cont_bits : std_logic_vector(5 downto 0);   -- numero do bit transmitido
    signal tx_words  : std_logic_vector(2 downto 0);   -- numero de palavras transmitidas
    signal start_cont_bits : std_logic;
    signal dec_cont_bits : std_logic;
    signal wr_tx_words : std_logic;
    signal dec_tx_words : std_logic;
    
    --
    -- registradores que so escritos a partir de uma CPU (array de *8* registradores)
    --
    type bank is array(0 to 5) of std_logic_vector(31 downto 0); 
    signal reg : bank ;                            
    signal start_tx : std_logic;

    signal ready : std_logic;  -- indica disponibilidade para iniciar uma transmisso

begin

    --
    -- PARTE 1: processo que permite a escrita por parte de uma entidade externa
	-- nos registradores internos ao perifrico
    --
    process(clock, reset)
    begin       
         if reset = '1' then
            reg <= (others=>(others=>'0'));
         elsif clock'event and clock='1' then   
            if ce='1' and rw='0' then
                reg( CONV_INTEGER(endereco)) <= data;
            end if;
        end if;
    end process;
    
    start_tx <= reg(5)(0); 
	-- bit '0' do registrador(5)  o sinal de controle que, em 1, ordena 
	-- o incio da transmisso. Obrigatoriamente,o reg 5 deve ser o ltimo
	-- escrito pela entidade externa.
 
 
    --
    -- PARTE 2: leitura por parte de uma entidade externa do sinal de
	-- disponibilidade para transmisso. Corresponde ao bit 0 no barramento de dados 
    --
    data <=  (0=>ready, others=>'0') when ce='1' and rw='1' else  (others=>'Z');

    --
    -- PARTE 3: mquina de estados de controle do perifrico (BLOCO DE CONTROLE)
    --
    process(clock, reset)
    begin       
       if reset = '1' then
            EA <= REP;
       elsif clock'event and clock='1' then  
            EA <= PE;     
       end  if;
    end process;

    process(EA, start_tx, ready, cont_bits, tx_words) 
	-- Ateno! Tudo que  testado em if ou case TEM que estar na lista de sensitividade!!
    begin      
	start_cont_bits <='0';
	dec_cont_bits <='0';
	wr_tx_words <='0';
	dec_tx_words <='0';
	linha <= '1';
	case EA is  
		when REP =>	
			if reset='1' then ready <='0'; end if; 
			-- Apenas para garantir valor inicial para ready no reset
			if start_tx='1' then
					ready <='0'; -- Ateno, mudado para manter ready em 1 at leitura pelo
								 -- processador ser feita!!	
					PE <= INICIO;
			else 	PE <= REP;
			end if;
		when INICIO =>
			start_cont_bits <='1';
			wr_tx_words <='1';
			if ready='0' then
					PE <= FIRST;
			elsif start_tx='1' then
					PE <= INICIO;
			else 	PE <= REP;
			end if;
		when FIRST =>
			dec_cont_bits <='1';
			linha <='0';
			PE <= TX;
		when TX =>
			linha <= reg(CONV_INTEGER(tx_words))(CONV_INTEGER(cont_bits));
			if cont_bits="000000" then
					PE <= LAST;
					dec_tx_words <='1';
			else 	dec_cont_bits <='1'; -- Aqui para evitar 1ue contador v abaixo de 0!
					PE <= TX;
			end if;
		when LAST =>
			linha <='1';
			if tx_words="000" then
					ready <='1';
					PE <= INICIO;
			else	start_cont_bits<='1';
					PE <= FIRST;
			end if;
	end case;

    end process;
        
    --
    -- PARTE 4: contadores controlados pela mquina de estados (BLOCO DE DADOS)
    --
	reg_tx_words:    process (reset, clock)	 -- contador de palavras
    begin
       if reset = '1' then
            tx_words <= (others=>'0');
       elsif clock'event and clock='1' then 
       	    if wr_tx_words='1' then
       	    	tx_words <= reg(0)(2 downto 0);
       	    elsif dec_tx_words='1' then
       	    	tx_words <= tx_words-1;
       	    end if; 	
       end  if;
    end process;
    
	reg_cont_bits:    process (reset, clock)  -- contador de bits
    begin
       if reset = '1' then
            cont_bits <= (others=>'0');
       elsif clock'event and clock='1' then 
       	    if start_cont_bits='1' then
       	    	cont_bits <= "100000";
       	    elsif dec_cont_bits ='1' then 
       	    	cont_bits <= cont_bits -1;
       	    end if;
       end  if;
    end process;
 
end transmissor;


-------------------------------------------------------------------------------------------------
-- a simple test bench
-------------------------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity transmissor_tb is
end transmissor_tb;

architecture TB_ARCHITECTURE of transmissor_tb is
  signal clock,reset: std_logic := '0';
  signal ce, rw, linha  : std_logic;
 
  signal endereco: std_logic_vector(2 downto 0);
  signal data:     std_logic_vector(31 downto 0);
  
  signal saida, verificacao: std_logic_vector(31 downto 0);
  signal bits_tx : integer;
 begin

  UUT : entity work.transmissor
          port map
            (clock => clock,  reset => reset, ce=>ce, rw=>rw,  
             endereco => endereco,  data => data, linha=>linha);

  -- gerador de clock de 100 MHz (10 ns de perodo)
  clock <= not clock after 5 ns;
            
  -- gerador do reset
  reset <= '1', '0' after 7 ns;
  
  process
  begin
  
     data     <= x"00000004";
     endereco <= "000";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;

     data     <= x"FFFFFFFF";
     endereco <= "001";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;          
     
     data     <= x"12345678";
     endereco <= "010";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;          
     
     data     <= x"9ABCDEF0";
     endereco <= "011";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;

     data     <= x"55555555";
     endereco <= "100";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;     

     data     <= x"00000001";
     endereco <= "101";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;  
     
     data     <= x"00000000" ;
     endereco <= "101";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;     
     
     ce       <= '0';
     rw       <= '0';	
	 wait for 1600ns; -- aguarda transmisso concluir  e comea a segunda transmissso
	 
     data     <= x"00000002";
     endereco <= "000";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;

     data     <= x"AAAAAAAA";
     endereco <= "001";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;          
     
     data     <= x"66666666";
     endereco <= "010";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;          
     
     data     <= x"00000001";
     endereco <= "101";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;  
     
     data     <= x"00000000" ;
     endereco <= "101";
     ce       <= '1';
     rw       <= '0';
     wait for 20 ns;     
     
     ce       <= '0';
     rw       <= '0';	
     wait until reset = '1' ;  -- tranca
     
   end process;
   
   -- PROCESSO DE VERIFICAO QUE PERMITE VERIFICAR SE O QUE SAI EM 'linha'  O QUE ESPERAVA ----
   process(clock, reset)
    begin       
       if reset = '1' then
           verificacao <= (others =>'0');
           bits_tx <= 0;
       elsif clock'event and clock='1' then 
           if bits_tx=0 and linha='0' then
               bits_tx<=1;
            elsif bits_tx>0 then
                verificacao <= linha & verificacao(31 downto 1);
                bits_tx<=bits_tx+1;
                if bits_tx=33 then
                  saida <= verificacao;
                  bits_tx <= 0;
                end if; 
                    end if;
       end if;
     end process;

end TB_ARCHITECTURE;
